/* vim:set ts=2 sw=2 sts=2 et: */
/**
 * \author     Marcus Holland-Moritz (github@mhxnet.de)
 * \copyright  Copyright (c) Marcus Holland-Moritz
 *
 * This file is part of dwarfs.
 *
 * dwarfs is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * dwarfs is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with dwarfs.  If not, see <https://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#pragma once

#include <string>
#include <string_view>

#include <dwarfs/writer/fragment_order_options.h>

namespace dwarfs {

class file_access;

namespace writer {

struct fragment_order_parser {
 public:
  static std::string choices();

  fragment_order_parser(std::shared_ptr<file_access const> const& fa)
      : fa_{fa} {}

  fragment_order_options parse(std::string_view arg) const;
  std::string to_string(fragment_order_options const& opts) const;

 private:
  std::shared_ptr<file_access const> fa_;
};

} // namespace writer
} // namespace dwarfs
