// autogenerated code, do not modify

#pragma once

// NOLINTBEGIN(cppcoreguidelines-macro-to-enum)
#define DWARFS_VERSION_MAJOR 0
#define DWARFS_VERSION_MINOR 15
#define DWARFS_VERSION_PATCH 6
#define DWARFS_VERSION (DWARFS_VERSION_MAJOR * 10000 + DWARFS_VERSION_MINOR * 100 + DWARFS_VERSION_PATCH)
// NOLINTEND(cppcoreguidelines-macro-to-enum)

namespace dwarfs {

int get_dwarfs_library_version();

extern char const* const DWARFS_GIT_REV;
extern char const* const DWARFS_GIT_DATE;
extern char const* const DWARFS_GIT_DESC;
extern char const* const DWARFS_GIT_BRANCH;
extern char const* const DWARFS_GIT_ID;

} // namespace dwarfs
